package org.insa.graphs.algorithm.shortestpath;

import static org.junit.Assert.*;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Random;

import org.insa.graphs.algorithm.AbstractInputData.Mode;
import org.insa.graphs.algorithm.ArcInspector;
import org.insa.graphs.algorithm.ArcInspectorFactory;
import org.insa.graphs.model.Graph;
import org.insa.graphs.model.Node;
import org.insa.graphs.model.Path;
import org.insa.graphs.model.io.BinaryGraphReader;
import org.insa.graphs.model.io.GraphReader;
import org.insa.graphs.model.io.PathReader;
import org.junit.BeforeClass;
import org.junit.Test;
import org.junit.runners.Parameterized.Parameter;

public abstract class ShortestPathAlgorithmTest {

    public static GraphReader reader;
    public static ArrayList<Graph> graph = new ArrayList<Graph>();
    public static PathReader pathReader;
    public static ArrayList<Path> path = new ArrayList<Path>();

    protected abstract ShortestPathAlgorithm initAlgo(ShortestPathData data);

    @Parameter
    // Liste de cartes utilisées
    static ArrayList<String> Maps = new ArrayList<String>(Arrays.asList("../Maps/carre.mapgr",
                                                                            "../Maps/insa.mapgr",
                                                                            "../Maps/toulouse.mapgr",
                                                                            "../Maps/midi-pyrenees.mapgr"));

    @BeforeClass
    /* 
     * Ajoute toute les cartes dans l'attribut graph.
     * Ensuite, on testera des chemins dont qui sont disponibles dans custom_path.
     * Ces chemins ont été tracés avec l'algo BellmanFord et sont donc considérés corrects.
     * On ne crée pas les chemins dans un attribut, on va les "recréer" virtuellement à
     * partir de leur origine/destination et en appliquant BellmanFord.
     * Cela permet une meilleure adaptabilité du code.
    */
    public static void init() {
        try {
            // Create the map
            for (String mapName : Maps) {
                // Create a graph reader
                final GraphReader reader = new BinaryGraphReader(
                    new DataInputStream(new BufferedInputStream(new FileInputStream(mapName))));
                // Read the graph.
                graph.add(reader.read());
                // free resources
                reader.close();
                }
            }
        catch (FileNotFoundException e) {
            System.err.println("File not found: " + e.getMessage());
            fail("File not found: " + e.getMessage());
        }
        catch (IOException e ) {
            System.err.println("Error reading file: " + e.getMessage());
            fail("Error reading file: " + e.getMessage());
        }
    }

    
    /* Stratégie:
     * Chemins courts testés par comparaison avec Bellman.
     * Chemin longs testés par comparaison avec chemins déjà construits. Bellman trop long.
     */

    /*
     * Verifies that path is valid and matches the one found by the Bellman algo
     * Otherwise the algo in not optimal.
     */
    private void assertBellmanHasSameResult(Graph graph, Node origin, Node destination, ArcInspector arcFilter) {
        final ShortestPathData data = new ShortestPathData(graph, origin, destination, arcFilter);

        final ShortestPathAlgorithm algo = initAlgo(data);
        final ShortestPathSolution path = algo.doRun();

        final BellmanFordAlgorithm bellman = new BellmanFordAlgorithm(data);
        final ShortestPathSolution bell_path = bellman.doRun();

        if (bell_path.getPath() != null) {
            if (bell_path.getPath().isValid()) {
                assert(path.getPath().isValid());
                assert(path.isFeasible());
                if (data.getMode() == Mode.LENGTH) {
                    assert(Math.abs((float) algo.getCostPath() - path.getPath().getLength()) < 0.1);
                    assert(Math.abs(path.getPath().getLength() - bell_path.getPath().getLength()) < 0.1);
                }
                else {
                    assert(Math.abs(algo.getCostPath() - path.getPath().getMinimumTravelTime()) < 0.1);
                    assert(Math.abs(path.getPath().getMinimumTravelTime() - bell_path.getPath().getMinimumTravelTime()) < 0.1);
                }
            }
            else {
                assert(!path.getPath().isValid());
                assert(!path.isFeasible());
            }
        }
        else {
            assert(!path.isFeasible());
        }
    }

    /*
     * Verifies that no path is found
     */
    private void assertNoPathFound(Graph graph, Node origin, Node destination, ArcInspector arcFilter) {
        ShortestPathData data = new ShortestPathData(graph, origin, destination, arcFilter);

        final ShortestPathAlgorithm algo = initAlgo(data);
        final ShortestPathSolution path = algo.doRun();
        
        assert(!path.isFeasible());
        assert(path.getPath() == null);
    }
    
    @Test
    /*
     * Map: carre.mapgr
     * Chemin: 19 --> 4
     * Tous chemins permis
     * Mode: LENGTH
     * PATH UTILISE : ../Paths/custom_paths/short_path_carre.path
    */
    public void chemin_court_CARRE_length() {
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(0); // all arcs
        Graph myGraph = graph.get(0);

        Node origin = myGraph.get(19);
        Node destination = myGraph.get(4);

        assertBellmanHasSameResult(myGraph, origin, destination, arcInspector);
    }

    @Test
    /*
    * Chemin long relativement à la carte carrée.
    * Chemin: 15 --> 9
    * Tous chemins permis
    * PATH UTILISE : ../Paths/custom_paths/long_path_carre.path
    */
    public void chemin_long_CARRE_length() {
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(0);
        Graph myGraph = graph.get(0);
        Node origin = myGraph.get(15);
        Node destination = myGraph.get(9);

        assertBellmanHasSameResult(myGraph, origin, destination, arcInspector);
    }


    @Test
    /*
     * Chemin nul sur carte carrée.
     * L'origine et la destination sont les mêmes (noeud 3).
     * Tous chemins permis
     */
    public void chemin_nul_CARRE() {
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(0);
        Graph myGraph = graph.get(0);
        Node origin = myGraph.get(3);
        Node destination = myGraph.get(3);

        assertNoPathFound(myGraph, origin, destination, arcInspector);
    }

    @Test
    /*
     * Chemin inexistant sur la carte INSA.
     * Origine: 224
     * Destination: 814.
     * Les 2 noeuds font partie de deux composantes non connexes.
     * Tous chemins permis.
    */
    public void chemin_inexistant_INSA() {
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(0);
        Graph myGraph = graph.get(1);
        Node origin = myGraph.get(224);
        Node destination = myGraph.get(814);

        assertNoPathFound(myGraph, origin, destination, arcInspector);
    }

    @Test
    /*
     * Chemin court sur la carte de Toulouse.
     * Origine : 8423 
     * Destination: 8435
     * Tous chemins permis.
     * PATH UTILISE : ../Paths/custom_paths/short_path_tls.path
     */
    public void chemin_court_TLS() {
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(0);
        Graph myGraph = graph.get(2);
        Node origin = myGraph.get(8423);
        Node destination = myGraph.get(8435);

        assertBellmanHasSameResult(myGraph, origin, destination, arcInspector);
    }

    @Test
    /*
     * Chemin long sur la carte de Toulouse.
     * Même si Bellman est de plus long à faire long à faire, ce test prend moins
     * de 3s, on estime que ce n'est pas trop et on va utiliser Bellman.
     * Par contre, dans le test sur la région midi_pyrenees qui arrive après, on va 
     * être obligé de trouver une autre solution.
     * Origine: 16644
     * Destination: 39229
     * Mode: LENGTH
     * PATH UTILISE : ../Paths/custom_paths/long_path_tls.path
     */
    public void chemin_long_TLS_length() {
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(0);
        Graph myGraph = graph.get(2);
        Node origin = myGraph.get(16644);
        Node destination = myGraph.get(39229);

        assertBellmanHasSameResult(myGraph, origin, destination, arcInspector);
    }


    @Test
    /*
     * Chemin long sur la carte de Toulouse.
     * Même si Bellman est de plus long à faire long à faire, ce test prend moins
     * de 3s, on estime que ce n'est pas trop et on va utiliser Bellman.
     * Par contre, dans le test sur la région midi_pyrenees qui arrive après, on va 
     * être obligé de trouver une autre solution.
     * Origine: 16644
     * Destination: 39229
     * Mode: TIME
     * PATH UTILISE : ../Paths/custom_paths/long_path_tls.path
     */
    public void chemin_long_TLS_time() {
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(2);
        Graph myGraph = graph.get(2);
        Node origin = myGraph.get(16644);
        Node destination = myGraph.get(39229);
        ShortestPathData data = new ShortestPathData(myGraph, origin, destination, arcInspector);
        
        final ShortestPathAlgorithm algo = initAlgo(data);
        final ShortestPathSolution path = algo.doRun();

        BellmanFordAlgorithm bellman = new BellmanFordAlgorithm(data);
        ShortestPathSolution bell_path = bellman.doRun();
        
        assert(path.getPath().isValid());
        assert(path.isFeasible());
        assert((Math.abs(algo.getCostPath() - path.getPath().getMinimumTravelTime()) < 1.0));
        assert(Math.abs(path.getPath().getMinimumTravelTime() - bell_path.getPath().getMinimumTravelTime()) < 1.0 );
    }

    @Test
    /*
     * Test du mode à vélo facultatif.
     * Nous prenons une origine sur l'autoroute et une destination en dehors.
     * Ce chemin est donc censé être utilisable en vélo mais pas en voiture.
     * Avec le filtre vélos, on obtient pas de chemin.
     * Avec le filtre voitures on obtient un chemin.
     * Origine: 19135
     * Destination: 1980
     * PATH UTILISE : ../Paths/custom_paths/path_cyclist.path
     */
    public void chemin_velo_uniquement() {
        // Filter: forBicyclesCustomT
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(4);
        
        Graph myGraph = graph.get(2);
        Node origin = myGraph.get(19135);
        Node destination = myGraph.get(1980);

        ShortestPathData data = new ShortestPathData(myGraph, origin, destination, arcInspector);
        
        ShortestPathAlgorithm algo_bicycle = initAlgo(data);
        ShortestPathSolution path_bicycle = algo_bicycle.doRun();
        
        BellmanFordAlgorithm bellman_bicycle = new BellmanFordAlgorithm(data);
        ShortestPathSolution bell_path_bicycle = bellman_bicycle.doRun();

        // Filter: forCarsL
        ArcInspector new_Inspector = ArcInspectorFactory.getAllFilters().get(1);
        data = new ShortestPathData(myGraph, origin, destination, new_Inspector);

        ShortestPathAlgorithm algo_car = initAlgo(data);
        ShortestPathSolution path_car = algo_car.doRun();

        assertEquals(path_bicycle.getPath(), null);
        assert(!path_bicycle.isFeasible());
        assertEquals(bell_path_bicycle.getPath(), null);
        assert(!bell_path_bicycle.isFeasible());
        
        assert(path_car.getPath() != null);
        assert(path_car.isFeasible());
    }

    @Test
    /*
     * On veut vérifier le bon fonctionnement des modes TIME et LENGTH. 
     * Pour cela, on va obtenir deux chemins avec l'algo de algo et vérifier
     * que:
     * -le chemin TIME est plus rapide en durée que le chemin LENGTH
     * -le chemin LENGTH est plus court en distance que le chemin LENGTH.
     * On prend un grand chemin pour être sur d'avoir une différence :
     * Origine: 16644
     * Destination: 39229
     * Mode: TIME/LENGTH
     * PATH UTILISE : ../Paths/custom_paths/long_path_tls.path
     */
    public void chemin_time_length_comparison() {
        Graph myGraph = graph.get(2);
        Node origin = myGraph.get(19135);
        Node destination = myGraph.get(1980);
        
        // Filter: forCarsL
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(1);
        ShortestPathData data = new ShortestPathData(myGraph, origin, destination, arcInspector);
        
        ShortestPathAlgorithm algo_L = initAlgo(data);
        ShortestPathSolution path_L = algo_L.doRun();

        // Filter: forCarsT
        ArcInspector new_Inspector = ArcInspectorFactory.getAllFilters().get(2);
        data = new ShortestPathData(myGraph, origin, destination, new_Inspector);

        ShortestPathAlgorithm algo_T = initAlgo(data);
        ShortestPathSolution path_T = algo_T.doRun();

        assert(path_L.getPath().isValid());
        assert(path_L.isFeasible());
        assert(path_T.getPath().isValid());
        assert(path_T.isFeasible());

        assert((Math.abs(algo_L.getCostPath() - path_L.getPath().getLength()) < 1.0));
        assert((Math.abs(algo_T.getCostPath() - path_T.getPath().getMinimumTravelTime()) < 1.0));

        assert(path_L.getPath().getLength() < path_T.getPath().getLength());
        assert(path_L.getPath().getMinimumTravelTime() > path_T.getPath().getMinimumTravelTime());
    }

    @Test
    /*
     * Cette fois-ci, Bellman est trop long à utiliser même une seule fois.
     * On va donc utiliser quelques techniques pour se rassurer.
     * -vérifier certains noeuds comme départ, destination, noeud pivot.
     * -vérifier le cout avec une estimation gentille.
     * -etc.
     * Origin: 279654
     * Destination: 481936
     * Mode: LENGTH
     * PATH UTILISE : ../Paths/custom_paths/long_chemin_midi_pyrenees.path
     */
    public void chemin_long_Midi_pyrenees_length() {
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(0);
        Graph myGraph = graph.get(3);
        Node origin = myGraph.get(279654);
        Node destination = myGraph.get(481936);
        ShortestPathData data = new ShortestPathData(myGraph, origin, destination, arcInspector);
        
        final ShortestPathAlgorithm algo = initAlgo(data);
        final ShortestPathSolution path = algo.doRun();

        assert(path.getPath().isValid());
        assert(path.isFeasible());
        // On a des erreurs d'arrondi assez grande avec la distance, mais elles sont mineures
        // relativement aux distance de 300000 ici.
        // Elles sont causées par le fait que les chemi
        assert((Math.abs(algo.getCostPath() - path.getPath().getLength())) < 1000.0);
        // Probable explication :
        // - `algo.getCostPath()` : somme de double
        // - `path.getPath().getLength()` : somme de float
        // Selon le chemin sélectionné on peut avoir une estimation de la longueur qu'on est censée avoir.
        // Avec notre long chemin: entre 250 et 260 kilomètres.
        assert(algo.getCostPath() > 250000 && algo.getCostPath() < 260000);
        // On peut aussi supposer que le nombre d'arcs empruntés est très grand.
        assert(path.getPath().getArcs().size() > 1000);
    }
    
    @Test
    /*
     * Cette fois-ci, Bellman est trop long à utiliser même une seule fois.
     * On va donc utiliser quelques techniques pour se rassurer.
     * -vérifier certains noeuds comme départ, destination, noeud pivot.
     * -vérifier le cout avec une estimation gentille.
     * -etc.
     * Origin: 279654
     * Destination: 481936
     * Mode: LENGTH
     * PATH UTILISE : ../Paths/custom_paths/long_chemin_midi_pyrenees.path
     */
    public void chemin_long_Midi_pyrenees_time() {
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(2);
        Graph myGraph = graph.get(3);
        Node origin = myGraph.get(279654);
        Node destination = myGraph.get(481936);
        ShortestPathData data = new ShortestPathData(myGraph, origin, destination, arcInspector);
        
        final ShortestPathAlgorithm algo = initAlgo(data);
        final ShortestPathSolution path = algo.doRun();

        assert(path.getPath().isValid());
        assert(path.isFeasible());
        // On a des erreurs d'arrondi assez grandes avec la distance
        assert((Math.abs(algo.getCostPath() - path.getPath().getMinimumTravelTime())) < 1000.0);
        // Selon le chemin sélectionné on peut avoir une estimation de la durée qu'on est censée avoir.
        // Avec notre long chemin: entre 12000 et 13000 secondes.
        assert(algo.getCostPath() > 12000 && algo.getCostPath() < 13000);
        // On peut aussi supposer que le nombre d'arcs empruntés est très grand.
        assert(path.getPath().getArcs().size() > 1000);
    }



@Test
    /*
     * Chemin long sur la carte de Toulouse.
     * Même si Bellman est de plus long à faire long à faire, ce test prend moins
     * de 3s, on estime que ce n'est pas trop et on va utiliser Bellman.
     * Par contre, dans le test sur la région midi_pyrenees qui arrive après, on va 
     * être obligé de trouver une autre solution.
     * Origine: 16644
     * Destination: 39229
     * Mode: LENGTH
     * PATH UTILISE : ../Paths/custom_paths/long_path_tls.path
     */
    public void cheminsToulouse() {
        ArcInspector arcInspector = ArcInspectorFactory.getAllFilters().get(0);
        Graph myGraph = graph.get(2);
        Node origin = myGraph.get(16644);
        Node destination = myGraph.get(39229);

        Random rand = new Random();

        int size_graph = myGraph.getNodes().size();

        for (int i = 0 ; i < 50 ; i++) {
            origin = myGraph.get(Math.abs(rand.nextInt()) % size_graph);
            destination = myGraph.get(Math.abs(rand.nextInt()) % size_graph);

            assertBellmanHasSameResult(myGraph, origin, destination, arcInspector);
        }
    }
}
